/*
 * TpHostGpu.cu
 *
 *  Created on: Apr 2, 2025
 *      Author: elphel
 */
#include <stdexcept>
#include <helper_cuda.h>  // for checkCudaErrors
#include <cuda_runtime.h> // cudaFree
#include <helper_functions.h> // timer functions

//#include "TpParams.h"   // TpHostGpu.h has it
#include "tp_paths.h"
#include "tp_files.h"
#include "tp_utils.h"       // for copyalloc_kernel_gpu
#include "TpHostGpu.h"


TpHostGpu::~TpHostGpu(){
	 hfree(m_host_kern_buf);
	 hfree(m_ftask_data);
	 hfree(m_ftask_data1);

	 hfree(m_correction_vector);
	 hfree(m_rByRDist);
	 gfree(m_gpu_geometry_correction);
	 gfree(m_gpu_correction_vector);
	 gfree(m_gpu_rByRDist);
	 gfree(m_gpu_rot_deriv);

	 for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
		 // Free used host memory, initialized as arrays of pointers
		 hfree(m_gpu_kernels_h[ncam]);
		 hfree(m_gpu_kernel_offsets_h[ncam]);
		 hfree(m_gpu_images_h[ncam]);
		 hfree(m_gpu_clt_h[ncam]);
		 hfree(m_gpu_corr_images_h[ncam]);
		 // Free used GPU memory initialized as **
		 if (m_gpu_kernels)          gfree(m_gpu_kernels[ncam]);
		 if (m_gpu_kernel_offsets)   gfree(m_gpu_kernel_offsets[ncam]);
		 if (m_gpu_images)           gfree(m_gpu_images[ncam]);
		 if (m_gpu_clt)              gfree(m_gpu_clt[ncam]);
		 if (m_gpu_corr_images)      gfree(m_gpu_corr_images[ncam]);
	 }
	 gfree(m_gpu_corrs);
	 gfree(m_gpu_corrs_td);
	 gfree(m_gpu_corrs_combo);
	 gfree(m_gpu_corrs_combo_td);
	 gfree(m_gpu_corr_indices);
	 gfree(m_gpu_corrs_combo_indices);
	 gfree(m_gpu_ftasks);
	 gfree(m_gpu_active_tiles);
	 gfree(m_gpu_num_active);
	 gfree(m_gpu_num_corr_tiles);
	 gfree(m_gpu_texture_indices);
	 gfree(m_gpu_woi);
	 gfree(m_gpu_twh);
	 gfree(m_gpu_num_texture_tiles);
	 gfree(m_gpu_port_offsets);
	 gfree(m_gpu_color_weights);
	 gfree(m_gpu_generate_RBGA_params);
	 gfree(m_gpu_textures);
	 gfree(m_gpu_diff_rgb_combo);
	 gfree(m_gpu_textures_rbga);

	 return;
};
void TpHostGpu::setImageKernels(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
       readFloatsFromFile(
       		m_host_kern_buf, // float * data, // allocated array
			m_tpPaths.kernel_file[ncam]); // 			   char *  path) // file path
       m_gpu_kernels_h[ncam] = copyalloc_kernel_gpu(m_host_kern_buf, m_tpParams.kern_size);
       readFloatsFromFile(
    		   m_host_kern_buf, // float * data, // allocated array
			   m_tpPaths.kernel_offs_file[ncam]); // 			   char *  path) // file path
       m_gpu_kernel_offsets_h[ncam] = (struct CltExtra *) copyalloc_kernel_gpu(
    		   m_host_kern_buf,
			   m_tpParams.kern_tiles * (sizeof( struct CltExtra)/sizeof(float)));
    }
    m_gpu_kernels =         copyalloc_pointers_gpu (m_gpu_kernels_h,     m_tpParams.num_cams); // NUM_CAMS);
    m_gpu_kernel_offsets = (struct CltExtra **) copyalloc_pointers_gpu ((float **) m_gpu_kernel_offsets_h, m_tpParams.num_cams); // NUM_CAMS);
}
void TpHostGpu::setCltBuffers(){
	const int slice_size {m_tpParams.tilesy * m_tpParams.tilesx * m_tpParams.num_colors * 4 * m_tpParams.dtt_size * m_tpParams.dtt_size};
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        // will get results back
    	//gpu_clt_h[ncam] = alloc_kernel_gpu(TILESY * TILESX * tpParams.num_colors * 4 * DTT_SIZE * DTT_SIZE);
        m_gpu_clt_h[ncam] = alloc_kernel_gpu(slice_size);
        printf("Allocating GPU memory, 0x%x floats\n", slice_size) ;
    }
    m_gpu_clt =            copyalloc_pointers_gpu (m_gpu_clt_h,         m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setCorrImgBuffers(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        // allocate result images (3x height to accommodate 3 colors
        // Image is extended by 4 pixels each side to avoid checking (mclt tiles extend by 4)
        //host array of pointers to GPU arrays
        m_gpu_corr_images_h[ncam] = alloc_image_gpu(
        		&dstride_rslt,                // size_t* dstride, // in bytes!!
				m_tpParams.img_width + m_tpParams.dtt_size,         // int width,
				m_tpParams.num_colors*(m_tpParams.img_height + m_tpParams.dtt_size));   // int height);
    }
    m_gpu_corr_images =    copyalloc_pointers_gpu (m_gpu_corr_images_h, m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setImgBuffers(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        readFloatsFromFile(
        		m_host_kern_buf, // float * data, // allocated array
				m_tpPaths.image_files[ncam]); // 			   char *  path) // file path
        m_gpu_images_h[ncam] =  copyalloc_image_gpu(
        		m_host_kern_buf, // float * image_host,
				&dstride,      // size_t* dstride,
				m_tpParams.img_width, // IMG_WIDTH,     // int width,
				m_tpParams.img_height); // IMG_HEIGHT);   // int height);
    }
    m_gpu_images =         copyalloc_pointers_gpu (m_gpu_images_h,      m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setImgBuffersShifted(int is_bayer, int image_dx, int image_dy) {
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        readFloatsFromFile(
        		m_host_kern_buf, // float * data, // allocated array
				m_tpPaths.image_files[ncam]); // 			   char *  path) // file path
        shift_image (
        		m_host_kern_buf,       // float * image,
				m_tpParams.img_width,  // int width,
				m_tpParams.img_height, // int height,
				is_bayer,              // int bayer,
				image_dx,              // int dx,
				image_dy);             // int dy);

        update_image_gpu(
        		m_host_kern_buf,          // float * image_host,
				m_gpu_images_h[ncam], // float * image_gpu,
				dstride,                // size_t  dstride, // in floats !
				m_tpParams.img_width, // IMG_WIDTH,     // int width,
				m_tpParams.img_height); // IMG_HEIGHT);   // int height);
        m_gpu_images_h[ncam] =  copyalloc_image_gpu(
        		m_host_kern_buf, // float * image_host,
				&dstride,      // size_t* dstride,
				m_tpParams.img_width, // IMG_WIDTH,     // int width,
				m_tpParams.img_height); // IMG_HEIGHT);   // int height);
    }
}



void TpHostGpu::setGeometryCorrectionBuffers() {
    readFloatsFromFile(
    		(float *) &m_fgeometry_correction, // float * data, // allocated array, no need to free
			m_tpPaths.geometry_correction_file); // char *  path) // file path
    m_rByRDist = readAllFloatsFromFile(
    		m_tpPaths.rByRDist_file, // const char *  path,
    		&m_rByRDist_length);       // int * len_in_floats)
    m_correction_vector =  readAllFloatsFromFile(
    		m_tpPaths.correction_vector_file, // const char *  path,
    		&m_correction_vector_length); // int * len_in_floats)
    m_gpu_geometry_correction =  (struct gc *) copyalloc_kernel_gpu(
    		(float *) &m_fgeometry_correction,
    		sizeof(m_fgeometry_correction)/sizeof(float));
    m_gpu_correction_vector =  (struct corr_vector * ) copyalloc_kernel_gpu(
    		m_correction_vector,
			m_correction_vector_length);
    m_gpu_rByRDist =  copyalloc_kernel_gpu(
    		m_rByRDist,
			m_rByRDist_length);
    checkCudaErrors(cudaMalloc((void **)&m_gpu_rot_deriv, sizeof(trot_deriv)));
}
void TpHostGpu::setCorrelationBuffers(){
    // allocates one correlation kernel per line (15x15 floats), number of rows - number of tiles * number of pairs
    m_gpu_corrs = alloc_image_gpu(
    		&dstride_corr,                  // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			m_tpParams.corr_length,                      // int width,
			m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy);   // int height);
    // read channel images (assuming host_kern_buf size > image size, reusing it)
    // allocate all other correlation data, some may be
    m_gpu_corrs_td = alloc_image_gpu(
    		&dstride_corr_td,                                                 // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			4 * m_tpParams.dtt_size *m_tpParams.dtt_size,                     // int width,
			m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy);    // int height);

    m_gpu_corrs_combo = alloc_image_gpu(
    		&dstride_corr_combo,                                              // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			m_tpParams.corr_length,                                             // int width,
			m_tpParams.tilesx * m_tpParams.tilesy);                           // int height);

    m_gpu_corrs_combo_td = alloc_image_gpu(
    		&dstride_corr_combo_td,                                           // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			4 * m_tpParams.dtt_size *m_tpParams.dtt_size,                     // int width,
			m_tpParams.tilesx * m_tpParams.tilesy);                           // int height);
    // just allocate
   checkCudaErrors (cudaMalloc((void **)&m_gpu_corr_indices,  m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy * sizeof(int)));
   checkCudaErrors (cudaMalloc((void **)&m_gpu_corrs_combo_indices,            m_tpParams.tilesx * m_tpParams.tilesy * sizeof(int)));
}

void TpHostGpu::setTasks(const float target_disparity, const float scale){
	// allocate m_ftask_data
	m_ftask_data  =   (float *) malloc( m_tpParams.tilesx *  m_tpParams.tilesy * m_tpParams.task_size * sizeof(float));
//    float * ftask_data1  =  (float *) malloc(TILESX * TILESY * tpParams.task_size * sizeof(float));

	   // tasks for all tiles
	    for (int ty = 0; ty < m_tpParams.tilesy ; ty++){
	        for (int tx = 0; tx < m_tpParams.tilesx ; tx++){
	            int nt = ty * m_tpParams.tilesx  + tx;
	            int task_task = (1 << m_tpParams.task_inter_en) | (1 << m_tpParams.task_corr_en) | (1 << m_tpParams.task_text_en); // just 1 bit, correlation selection is defined by common corr_sel bits
	            int task_txy = tx + (ty << 16);
	//            float task_target_disparity = DBG_DISPARITY; // disparity for which to calculate offsets (not needed in Java)
	            float * tp = m_ftask_data + m_tpParams.task_size * nt;
	            *(tp + m_tpParams.tp_task_task_offset) =    *(float *) &task_task;
	            *(tp + m_tpParams.tp_task_txy_offset) =      *(float *) &task_txy;
	            *(tp + m_tpParams.tp_task_disparity_offset) =  target_disparity;
	//            tp += 2; // skip centerX, centerY
	            *(tp + m_tpParams.tp_task_scale_offset) = scale; //  0; // 0.5f; // ,0; // scale, 0 - old way, just set
	            tp+= m_tpParams.tp_task_xy_offset;
	            for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
	            	*(tp++) = m_tile_coords_h[ncam][nt][0];
	            	*(tp++) = m_tile_coords_h[ncam][nt][1];
	            }
	        }
	    }
	    m_gpu_ftasks = (float  *) copyalloc_kernel_gpu(m_ftask_data, m_tpParams.tp_tasks_size * m_tpParams.task_size); // (sizeof(struct tp_task)/sizeof(float)));
}
void TpHostGpu::setTextures(){
	if (!m_ftask_data) { // ftask_data is not initialized
		throw std::runtime_error("Error: m_ftask_data is not initialized ");
	}
	m_tile_texture_layers = (m_tpParams.texture_colors + 1 + (m_tpParams.keep_texture_weights? (m_tpParams.num_cams + m_tpParams.texture_colors + 1): 0));
	m_tile_texture_size = m_tile_texture_layers *256;
	 // task data should be initialized
	m_num_textures=0;
    for (int ty = 0; ty < m_tpParams.tilesy; ty++){
    	for (int tx = 0; tx < m_tpParams.tilesx; tx++){
    		int nt = ty * m_tpParams.tilesx + tx;
            float *tp = m_ftask_data + m_tpParams.task_size * nt;
    		int cm = (*(int *) tp) & m_tpParams.task_texture_bits; // non-zero any of 4 lower task bits
    		if (cm){
    			m_texture_indices[m_num_textures++] = (nt << m_tpParams.text_ntile_shift) | (1 << m_tpParams.list_texture_bit); // setting  0x80 in texture indices
    		}
    	}
    }
    // num_textures now has the total number of textures
    // copy corr_indices to gpu
    m_gpu_texture_indices = (int  *) copyalloc_kernel_gpu(
    		(float * ) m_texture_indices,
			m_num_textures,
			m_tpParams.tilesx * m_tpParams.tilesya); // number of rows - multiple of 4
    m_gpu_textures = alloc_image_gpu(
    		&dstride_textures,                      // in bytes ! for one rgba/ya 16x16 tile
			m_tile_texture_size,                    // int width (floats),
			m_tpParams.tilesx * m_tpParams.tilesy); // int height);
}

void TpHostGpu::setRGBA(){
	/*
	// appears in setTextures
    m_tile_texture_layers = (m_tpParams.texture_colors + 1 + (m_tpParams.keep_texture_weights? (m_tpParams.num_cams + m_tpParams.texture_colors + 1): 0));
    m_tile_texture_size = m_tile_texture_layers *256;
    */
    m_rgba_width =   (m_tpParams.tilesx + 1) * m_tpParams.dtt_size;
    m_rgba_height =  (m_tpParams.tilesy + 1) * m_tpParams.dtt_size;;
    m_rbga_slices =  m_tpParams.texture_colors + 1; // 4/1
    if (m_tpParams.keep_texture_weights & 2){
    	m_rbga_slices += m_tpParams.texture_colors * m_tpParams.num_cams;
    }
    checkCudaErrors(cudaMalloc((void **)&m_gpu_woi,               4 * sizeof(float))); // RGBA
    checkCudaErrors(cudaMalloc((void **)&m_gpu_twh,               2 * sizeof(float))); // RGBA, dynamic
    checkCudaErrors(cudaMalloc((void **)&m_gpu_num_texture_tiles, 8 * sizeof(float))); // for each subsequence - number of non-border, RGBA all
	m_gpu_port_offsets =         (float *) copyalloc_kernel_gpu((float * ) m_tpParams.port_offsets, m_tpParams.num_cams * 2); // num_ports * 2);
    m_gpu_color_weights =        (float *) copyalloc_kernel_gpu((float * ) m_tpParams.color_weights, sizeof(m_tpParams.color_weights));
    m_gpu_generate_RBGA_params = (float *) copyalloc_kernel_gpu((float * ) m_tpParams.generate_RBGA_params, sizeof(m_tpParams.generate_RBGA_params));

    m_gpu_textures_rbga = alloc_image_gpu(
   		&dstride_textures_rbga,              // in bytes ! for one rgba/ya 16x16 tile
			m_rgba_width,              // int width (floats),
			m_rgba_height * m_rbga_slices);               // int height);
    checkCudaErrors(cudaMalloc((void **)&m_gpu_diff_rgb_combo,  m_tpParams.tilesx * m_tpParams.tilesy * m_tpParams.num_cams *
    		(m_tpParams.num_colors + 1) * sizeof(float)));
}

trot_deriv TpHostGpu::testRotMatrices (int num_runs){ // 424
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_correction_vector){
		throw std::runtime_error("Error: m_gpu_correction_vector is not initialized ");
	}
	if (!m_gpu_rot_deriv){
		throw std::runtime_error("Error: m_gpu_rot_deriv is not initialized ");
	}
    dim3 threads_rot (3,3,3);
    dim3 grid_rot    (m_tpParams.num_cams, 1, 1);
    printf("ROT_MATRICES: threads_list=(%d, %d, %d)\n",threads_rot.x,threads_rot.y,threads_rot.z);
    printf("ROT_MATRICES: grid_list=(%d, %d, %d)\n",grid_rot.x,grid_rot.y,grid_rot.z);
    StopWatchInterface *timerROT_MATRICES = 0;
    sdkCreateTimer(&timerROT_MATRICES);
    for (int i = i0; i < numIterations; i++) {
    	if (i == 0)	{
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerROT_MATRICES);
    		sdkStartTimer(&timerROT_MATRICES);
    	}
    	calc_rot_deriv<<<grid_rot,threads_rot>>> (
    			m_tpParams.num_cams,                // int                  num_cams,
				m_gpu_correction_vector ,  // struct corr_vector * gpu_correction_vector,
				m_gpu_rot_deriv);          // union trot_deriv   * gpu_rot_deriv);
    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("testRotMatrices pass: %d\n",i);
    }
    ///	cudaProfilerStop();
    sdkStopTimer(&timerROT_MATRICES);
    float avgTimeROT_MATRICES = (float)sdkGetTimerValue(&timerROT_MATRICES) / (float)numIterations;
    sdkDeleteTimer(&timerROT_MATRICES);
    printf("Average calc_rot_matrices run time =%f ms\n",  avgTimeROT_MATRICES);
    trot_deriv rot_deriv{};
	checkCudaErrors(cudaMemcpy(
			&rot_deriv,
			m_gpu_rot_deriv,
			sizeof(trot_deriv),
			cudaMemcpyDeviceToHost));
	return rot_deriv;
}

void TpHostGpu::testReverseDistortions (int num_runs){ // 468
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_geometry_correction){
		throw std::runtime_error("Error: m_gpu_geometry_correction is not initialized ");
	}
	if (!m_gpu_rByRDist){
		throw std::runtime_error("Error: m_gpu_rByRDist is not initialized ");
	}
	if (!m_rByRDist){
		throw std::runtime_error("Error: m_rByRDist is not initialized ");
	}
    dim3 threads_rd(3,3,3);
    dim3 grid_rd   (NUM_CAMS, 1, 1); // can get rid of NUM_CAMS
    printf("REVERSE DISTORTIONS: threads_list=(%d, %d, %d)\n",threads_rd.x,threads_rd.y,threads_rd.z);
    printf("REVERSE DISTORTIONS: grid_list=(%d, %d, %d)\n",grid_rd.x,grid_rd.y,grid_rd.z);
    StopWatchInterface *timerREVERSE_DISTORTIONS = 0;
    sdkCreateTimer(&timerREVERSE_DISTORTIONS);
    for (int i = i0; i < numIterations; i++) {
    	if (i == 0) {
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerREVERSE_DISTORTIONS);
    		sdkStartTimer(&timerREVERSE_DISTORTIONS);
    	}
    	calcReverseDistortionTable<<<grid_rd,threads_rd>>>(
    			m_gpu_geometry_correction, // 		struct gc          * gpu_geometry_correction,
				m_gpu_rByRDist);
    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("testReverseDistortions pass: %d\n",i);
    }
    ///	cudaProfilerStop();
    sdkStopTimer(&timerREVERSE_DISTORTIONS);
    float avgTimeREVERSE_DISTORTIONS = (float)sdkGetTimerValue(&timerREVERSE_DISTORTIONS) / (float)numIterations;
    sdkDeleteTimer(&timerREVERSE_DISTORTIONS);
    printf("Average calcReverseDistortionTable  run time =%f ms\n",  avgTimeREVERSE_DISTORTIONS);
    float * rByRDist_gen = (float *) malloc(m_tpParams.rbyrdist_len * sizeof(float));
	checkCudaErrors(cudaMemcpy(
			rByRDist_gen,
			m_gpu_rByRDist,
			m_tpParams.rbyrdist_len * sizeof(float),
			cudaMemcpyDeviceToHost));
	float max_err = 0;
	for (int i = 0; i < m_tpParams.rbyrdist_len; i++){
		float err = abs(rByRDist_gen[i] - m_rByRDist[i]);
		if (err > max_err){
			max_err = err;
		}
	}
	printf("Maximal rByRDist error = %f\n",max_err);
	free (rByRDist_gen);
    // temporarily restore
	if (0) {
		checkCudaErrors(cudaMemcpy(
				m_gpu_rByRDist,
				m_rByRDist,
				m_tpParams.rbyrdist_len * sizeof(float),
				cudaMemcpyHostToDevice));
	}
}

void TpHostGpu::testGeomCorrect         (int num_runs){ // 534
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_ftasks){
		throw std::runtime_error("Error: m_gpu_ftasks is not initialized ");
	}
	if (!m_gpu_geometry_correction){
		throw std::runtime_error("Error: m_gpu_geometry_correction is not initialized ");
	}
	if (!m_gpu_correction_vector){
		throw std::runtime_error("Error: m_gpu_correction_vector is not initialized ");
	}
	if (!m_gpu_rByRDist){
		throw std::runtime_error("Error: m_gpu_rByRDist is not initialized ");
	}
	if (!m_gpu_rot_deriv){
		throw std::runtime_error("Error: m_gpu_rot_deriv is not initialized ");
	}
	if (!m_ftask_data){
		throw std::runtime_error("Error: m_ftask_data is not initialized ");
	}
	if (!m_ftask_data1){
		throw std::runtime_error("Error: m_ftask_data1 is not initialized ");
	}
    dim3 threads_geom(m_tpParams.num_cams, m_tpParams.tiles_per_block_geom, 1);
    dim3 grid_geom   ((m_tpParams.tp_tasks_size + m_tpParams.tiles_per_block_geom-1)/m_tpParams.tiles_per_block_geom, 1, 1);
    printf("GEOM: threads_list=(%d, %d, %d)\n",threads_geom.x,threads_geom.y,threads_geom.z);
    printf("GEOM: grid_list=(%d, %d, %d)\n",grid_geom.x,grid_geom.y,grid_geom.z);
    StopWatchInterface *timerGEOM = 0;
    sdkCreateTimer(&timerGEOM);
    for (int i = i0; i < numIterations; i++) {
    	if (i == 0) {
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerGEOM);
    		sdkStartTimer(&timerGEOM);
    	}
    	calculate_tiles_offsets<<<1,1>>> (
    			1,                         // int                  uniform_grid, //==0: use provided centers (as for interscene) , !=0 calculate uniform grid
				m_tpParams.num_cams,                // int                  num_cams,
				m_gpu_ftasks,              // float              * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
//    			gpu_tasks,                 // struct tp_task     * gpu_tasks,
				m_tpParams.tp_tasks_size,  // int                  num_tiles,          // number of tiles in task list
				m_gpu_geometry_correction, //	struct gc          * gpu_geometry_correction,
				m_gpu_correction_vector,   //	struct corr_vector * gpu_correction_vector,
				m_gpu_rByRDist,            //	float *              gpu_rByRDist)      // length should match RBYRDIST_LEN
				m_gpu_rot_deriv);          // union trot_deriv   * gpu_rot_deriv);

    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("test pass: %d\n",i);
    }
    ///	cudaProfilerStop();
    sdkStopTimer(&timerGEOM);
    float avgTimeGEOM = (float)sdkGetTimerValue(&timerGEOM) / (float)numIterations;
    sdkDeleteTimer(&timerGEOM);
    printf("Average TextureList run time =%f ms\n",  avgTimeGEOM);

    checkCudaErrors(cudaMemcpy( // copy modified/calculated tasks
    		m_ftask_data1,
			m_gpu_ftasks,
			m_tpParams.tp_tasks_size * m_tpParams.task_size *sizeof(float),
			cudaMemcpyDeviceToHost));
    //task_size
#if 0 // for manual browsing
    struct tp_task * old_task = &task_data [DBG_TILE];
    struct tp_task * new_task = &task_data1[DBG_TILE];
#endif
	if( m_tpParams.debug_tile) {
	    printf("old_task txy = 0x%x\n", *(int *) (m_ftask_data + m_tpParams.tp_tasks_size * m_tpParams.dbg_tile + 1)) ; // task_data [DBG_TILE].txy);
	    printf("new_task txy = 0x%x\n", *(int *) (m_ftask_data1 + m_tpParams.task_size * m_tpParams.dbg_tile + 1)) ; // task_data1[DBG_TILE].txy);
	        for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++){
	            printf("camera %d pX old %f new %f diff = %f\n", ncam,
	            		 *(m_ftask_data  + m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 0),
	            		 *(m_ftask_data1 + m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 0),
	            		 (*(m_ftask_data + m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 0)) -
	            		 (*(m_ftask_data1 + m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 0)));
	            printf("camera %d pY old %f new %f diff = %f\n", ncam,
	           		 *(m_ftask_data  +  m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 1),
	           		 *(m_ftask_data1 +  m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 1),
	           		 (*(m_ftask_data +  m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 1)) -
	           		 (*(m_ftask_data1 + m_tpParams.task_size * m_tpParams.dbg_tile + m_tpParams.tp_task_xy_offset + 2*ncam + 1)));
	        }
	}
}

void TpHostGpu::testConvertDirect       (int num_runs){ // 608
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_kernel_offsets){
		throw std::runtime_error("Error: m_gpu_kernel_offsets is not initialized ");
	}
	if (!m_gpu_kernels){
		throw std::runtime_error("Error: m_gpu_kernels is not initialized ");
	}
	if (!m_gpu_images){
		throw std::runtime_error("Error: m_gpu_images is not initialized ");
	}
	if (!m_gpu_ftasks){
		throw std::runtime_error("Error: m_gpu_ftasks is not initialized ");
	}
	if (!m_gpu_clt){
		throw std::runtime_error("Error: m_gpu_clt is not initialized ");
	}
	if (!m_gpu_active_tiles){
		throw std::runtime_error("Error: m_gpu_active_tiles is not initialized ");
	}
	if (!m_gpu_num_active){
		throw std::runtime_error("Error: m_gpu_num_active is not initialized ");
	}
    //create and start CUDA timer
    StopWatchInterface *timerTP = 0;
    sdkCreateTimer(&timerTP);
    dim3 threads_tp(1, 1, 1);
    dim3 grid_tp(1, 1, 1);
    printf("threads_tp=(%d, %d, %d)\n",threads_tp.x,threads_tp.y,threads_tp.z);
    printf("grid_tp=   (%d, %d, %d)\n",grid_tp.x,   grid_tp.y,   grid_tp.z);
    ///    cudaProfilerStart();
    float ** fgpu_kernel_offsets = (float **) m_gpu_kernel_offsets; //  [tpParams.num_cams]  [NUM_CAMS];
    for (int i = i0; i < numIterations; i++) {
    	if (i == 0) {
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerTP);
    		sdkStartTimer(&timerTP);
    	}
    	convert_direct<<<grid_tp,threads_tp>>>( // called with a single block, CONVERT_DIRECT_INDEXING_THREADS threads
    			m_tpParams.num_cams,              // int                num_cams,           // actual number of cameras
				m_tpParams.num_colors,            // int                num_colors,         // actual number of colors: 3 for RGB, 1 for LWIR/mono
    			fgpu_kernel_offsets,     // struct CltExtra ** gpu_kernel_offsets,
				m_gpu_kernels,           // float           ** gpu_kernels,
				m_gpu_images,            // float           ** gpu_images,
				m_gpu_ftasks,            // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
				m_gpu_clt,               // float           ** gpu_clt,            // [num_cams][TILESY][TILESX][num_colors][DTT_SIZE*DTT_SIZE]
				dstride/sizeof(float),   // size_t             dstride, // for gpu_images
				m_tpParams.tp_tasks_size,// int                num_tiles) // number of tiles in task
				0,                       // int                lpf_mask)            // apply lpf to colors : bit 0 - red, bit 1 - blue, bit2 - green
				m_tpParams.img_width,    // IMG_WIDTH,               // int                woi_width,
				m_tpParams.img_height,   // IMG_HEIGHT,              // int                woi_height,
				0, // m_tpParams.kernels_hor, // KERNELS_HOR,           // int                kernels_hor,
				m_tpParams.kernels_hor,  // KERNELS_VERT,          // int                kernels_vert);
				m_gpu_active_tiles,      // int *              gpu_active_tiles,      // pointer to the calculated number of non-zero tiles
    			m_gpu_num_active, //);       // int *              pnum_active_tiles);  //  indices to gpu_tasks
				m_tpParams.tilesx); // TILESX); // int                tilesx)
        printf("HOST: convert_direct() done\n");
    	getLastCudaError("Kernel execution failed");
        printf("HOST: convert_direct() done - 1\n");
    	checkCudaErrors(cudaDeviceSynchronize());
        printf("HOST: convert_direct() done - 2\n");
//    	printf("%d\n",i);
    }
    sdkStopTimer(&timerTP);
    float avgTime = (float)sdkGetTimerValue(&timerTP) / (float)numIterations;
    sdkDeleteTimer(&timerTP);
    int num_active_tiles; // calculated by convert_direct
	checkCudaErrors(cudaMemcpy(
			&num_active_tiles,
			m_gpu_num_active, // make it local?
			sizeof(int),
			cudaMemcpyDeviceToHost));
    printf("Run time =%f ms, num active tiles = %d\n",  avgTime, num_active_tiles);
    saveClt(
    		m_tpPaths.ports_clt_file, // const char ** paths,    // tpPaths.ports_clt_file
			"CLT data",               // const char * prompt,    // "CLT data"
			m_gpu_clt_h);             // float **     gpu_clt_h);
}

void TpHostGpu::saveClt(
		const char ** paths,    // tpPaths.ports_clt_file
		const char * prompt,    // "CLT data"
		float **     gpu_clt_h){ // m_gpu_clt_h
	if (!paths)	return;
    int rslt_size = (m_tpParams.tilesy * m_tpParams.tilesx * m_tpParams.num_colors * 4 * m_tpParams.dtt_size * m_tpParams.dtt_size);
    float * cpu_clt = (float *)malloc(rslt_size*sizeof(float));
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) if (paths[ncam] && m_gpu_clt_h[ncam]){
    	checkCudaErrors(cudaMemcpy( // segfault
    			cpu_clt,
				m_gpu_clt_h[ncam],
				rslt_size * sizeof(float),
    			cudaMemcpyDeviceToHost));
        printf("Writing %s to %s\n",  prompt, paths[ncam]);
    	writeFloatsToFile(cpu_clt, // float *       data, // allocated array
    			rslt_size, // int           size, // length in elements
				paths[ncam]); // 			   const char *  path) // file path
    }
    hfree(cpu_clt);
}

void TpHostGpu::saveRgb(
		const char ** paths,  // m_tpPaths.result_rbg_file
		const char * prompt,  // "RBG data"
		float **     gpu_corr_images_h){
	if (!paths)	return;
	int rslt_img_size =       m_tpParams.num_colors * (m_tpParams.img_height + m_tpParams.dtt_size) * (m_tpParams.img_width + m_tpParams.dtt_size);
	float * cpu_corr_image = (float *)malloc(rslt_img_size * sizeof(float));
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) if (paths[ncam] && gpu_corr_images_h[ncam]){
    	checkCudaErrors(cudaMemcpy2D( // segfault
    			cpu_corr_image,
				(m_tpParams.img_width + m_tpParams.dtt_size) * sizeof(float),
				gpu_corr_images_h[ncam],
				dstride_rslt,
				(m_tpParams.img_width + m_tpParams.dtt_size) * sizeof(float),
				m_tpParams.num_colors* (m_tpParams.img_height + m_tpParams.dtt_size),
    			cudaMemcpyDeviceToHost));
        printf("Writing %s to %s\n",  prompt, paths[ncam]);
    	writeFloatsToFile( // will have margins
    			cpu_corr_image, // float *       data, // allocated array
				rslt_img_size,  // int           size, // length in elements
				paths[ncam]);   // 			   const char *  path) // file path
    }
    free(cpu_corr_image);
}

/*
// not implemented
void TpHostGpu::testImclt               (int num_runs){ // 682
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_geometry_correction){
		throw std::runtime_error("Error: m_gpu_geometry_correction is not initialized ");
	}
}
*/
void TpHostGpu::testImcltRbgAll         (int num_runs){ // 701
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_clt){
		throw std::runtime_error("Error: m_gpu_clt is not initialized ");
	}
	if (!m_gpu_corr_images){
		throw std::runtime_error("Error: m_gpu_corr_images is not initialized ");
	}
    StopWatchInterface *timerIMCLT = 0;
    sdkCreateTimer(&timerIMCLT);
    for (int i = i0; i < numIterations; i++) {
    	if (i == 0) {
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerIMCLT);
    		sdkStartTimer(&timerIMCLT);
    	}
        dim3 threads_imclt_all(1, 1, 1);
		dim3 grid_imclt_all(1, 1, 1);
        printf("threads_imclt_all=(%d, %d, %d)\n",threads_imclt_all.x,threads_imclt_all.y,threads_imclt_all.z);
        printf("grid_imclt_all=   (%d, %d, %d)\n",grid_imclt_all.x,   grid_imclt_all.y,   grid_imclt_all.z);
        imclt_rbg_all<<<grid_imclt_all,threads_imclt_all>>>(
        		m_tpParams.num_cams,                    // int                num_cams,
        		m_gpu_clt,                     // float           ** gpu_clt,            // [num_cams][TILESY][TILESX][num_colors][DTT_SIZE*DTT_SIZE]
				m_gpu_corr_images,             // float           ** gpu_corr_images,    // [num_cams][WIDTH, 3 * HEIGHT]
				1,                           // int               apply_lpf,
				m_tpParams.num_colors,                  // int               colors,               // defines lpf filter
				m_tpParams.tilesx,      // TILESX,                      // int               woi_twidth,
				m_tpParams.tilesy,      // TILESY,                      // int               woi_theight,
				dstride_rslt/sizeof(float)); // const size_t      dstride);            // in floats (pixels)
    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("test pass: %d\n",i);
    }
    // TODO: *** Stop here for initial testing ***
    sdkStopTimer(&timerIMCLT);
    float avgTimeIMCLT = (float)sdkGetTimerValue(&timerIMCLT) / (float)numIterations;
    sdkDeleteTimer(&timerIMCLT);
    printf("Average imclt_rbg_all run time =%f ms\n",  avgTimeIMCLT);
    saveRgb(
    		m_tpPaths.result_rbg_file, // const char ** paths,  // m_tpPaths.result_rbg_file
    		"RBG data",                // const char * prompt,  // "RBG data"
    		m_gpu_corr_images_h); // float **     gpu_corr_images_h){
}

void TpHostGpu::testCorrelate2DIntra(int num_runs){
	int  num_corr_indices = m_tpParams.num_pairs * m_tpParams.num_tiles;
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_clt){
		throw std::runtime_error("Error: m_gpu_clt is not initialized ");
	}
	if (!m_gpu_ftasks) {
		throw std::runtime_error("Error: m_gpu_ftasks is not initialized ");
	}
	if (!m_gpu_corrs) {
		throw std::runtime_error("Error: m_gpu_corrs is not initialized ");
	}
    StopWatchInterface *timerCORR = 0;
    sdkCreateTimer(&timerCORR);
    for (int i = i0; i < numIterations; i++)
    {
    	if (i == 0)
    	{
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerCORR);
    		sdkStartTimer(&timerCORR);
    	}
    	correlate2D<<<1,1>>>(
    			m_tpParams.num_cams,         // int               num_cams,
				m_tpParams.sel_pairs[0],     // int               sel_pairs0           // unused bits should be 0
				m_tpParams.sel_pairs[1],     // int               sel_pairs1,           // unused bits should be 0
				m_tpParams.sel_pairs[2],     // int               sel_pairs2,           // unused bits should be 0
				m_tpParams.sel_pairs[3],     // int               sel_pairs3,           // unused bits should be 0
				m_gpu_clt,                   // float          ** gpu_clt,            // [NUM_CAMS] ->[TILESY][TILESX][NUM_COLORS][DTT_SIZE*DTT_SIZE]
				m_tpParams.num_colors,       // int               colors,             // number of colors (3/1)
				m_tpParams.color_weights[0], // 0.25,  // float             scale0,             // scale for R
				m_tpParams.color_weights[1], // 0.25,  // float             scale1,             // scale for B
				m_tpParams.color_weights[2], // 0.5,   // float             scale2,             // scale for G
				m_tpParams.fat_zero * m_tpParams.fat_zero,                // float             fat_zero2,           // here - absolute
				m_gpu_ftasks,                // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
				m_tpParams.tp_tasks_size,    // int               num_tiles) // number of tiles in task
				m_tpParams.tilesx,           // int               tilesx,             // number of tile rows
				m_gpu_corr_indices,          // int             * gpu_corr_indices,   // packed tile+pair
				m_gpu_num_corr_tiles,        // int             * pnum_corr_tiles,    // pointer to a number of correlation tiles to process
				dstride_corr/sizeof(float),  // const size_t      corr_stride,        // in floats
				m_tpParams.corr_out_rad,     // int               corr_radius,        // radius of the output correlation (7 for 15x15)
				m_gpu_corrs);                // float           * gpu_corrs);          // correlation output data

    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("test pass: %d\n",i);
    }
    float avgTimeCORR = (float)sdkGetTimerValue(&timerCORR) / (float)numIterations;
    sdkStopTimer(&timerCORR);
    sdkDeleteTimer(&timerCORR);
    int num_corrs{}; // will get data from the gpu memory
    checkCudaErrors(cudaMemcpy(
    		&num_corrs,
			m_gpu_num_corr_tiles,
			sizeof(int),
			cudaMemcpyDeviceToHost));
    printf("Average CORR run time =%f ms, num cor tiles (new) = %d\n",  avgTimeCORR, num_corrs);
    saveIntraCorrFile(
    		m_tpPaths.result_corr_file, // const char * path,
			"phase correlation data",   // const char * prompt,
			num_corrs,                  // int          num_corrs,
			num_corr_indices,           // int          num_corr_indices,
			m_gpu_corrs,                // float *      gpu_corrs,
			m_gpu_corr_indices,         // int *        gpu_corr_indices)
			16);                        //int          num_sel_sensors) { // only for interscene

}

void TpHostGpu::testCorrelate2DInterSelf(int num_runs){
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	// check/replace names
	if (!m_gpu_clt){
		throw std::runtime_error("Error: m_gpu_clt is not initialized ");
	}
	if (!m_gpu_ftasks) {
		throw std::runtime_error("Error: m_gpu_ftasks is not initialized ");
	}
	if (!m_gpu_corrs) {
		throw std::runtime_error("Error: m_gpu_corrs is not initialized ");
	}
    int sel_sensors =      0xffff; // 0x7fff; // 0xffff;
    int num_sel_sensors =  16; // 15; // 16;
    int num_pairs_inter =  num_sel_sensors+1;
    int num_corr_indices = num_pairs_inter * m_tpParams.num_tiles;
    int is_bayer =         0;
    int image_dx =         2;
    int image_dy =         0;
    float            * gpu_clt_ref_h            [m_tpParams.num_cams];
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
       gpu_clt_ref_h[ncam] = alloc_kernel_gpu(m_tpParams.tilesy * m_tpParams.tilesx * m_tpParams.num_colors * 4 * m_tpParams.dtt_size * m_tpParams.dtt_size);
    }
    float           ** gpu_clt_ref =      copyalloc_pointers_gpu (gpu_clt_ref_h,  m_tpParams.num_cams); // NUM_CAMS);
    dim3 threads_tp(1, 1, 1);
    dim3 grid_tp(1, 1, 1);
    float ** fgpu_kernel_offsets = (float **) m_gpu_kernel_offsets; //  [tpParams.num_cams]  [NUM_CAMS];

    // use gpu_images and convert to gpu_clt_ref
	convert_direct<<<grid_tp,threads_tp>>>( // called with a single block, CONVERT_DIRECT_INDEXING_THREADS threads
			m_tpParams.num_cams,              // int                num_cams,           // actual number of cameras
			m_tpParams.num_colors,            // int                num_colors,         // actual number of colors: 3 for RGB, 1 for LWIR/mono
			fgpu_kernel_offsets,   // struct CltExtra ** gpu_kernel_offsets,
			m_gpu_kernels,           // float           ** gpu_kernels,
			m_gpu_images,            // float           ** gpu_images,
			m_gpu_ftasks,            // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
			gpu_clt_ref, //******    // float           ** gpu_clt,            // [num_cams][TILESY][TILESX][num_colors][DTT_SIZE*DTT_SIZE]
			dstride/sizeof(float),   // size_t             dstride, // for gpu_images
			m_tpParams.tp_tasks_size,// int                num_tiles) // number of tiles in task
			0,                       // int                lpf_mask)            // apply lpf to colors : bit 0 - red, bit 1 - blue, bit2 - green
			m_tpParams.img_width,    // int                woi_width,
			m_tpParams.img_height,   // int                woi_height,
			m_tpParams.kernels_hor,  // int                kernels_hor,
			m_tpParams.kernels_vert, //,          // int                kernels_vert);
			m_gpu_active_tiles,      // int *              gpu_active_tiles,      // pointer to the calculated number of non-zero tiles
			m_gpu_num_active,        // int *              pnum_active_tiles);  //  indices to gpu_tasks
			m_tpParams.tilesx);      // int                tilesx)
	getLastCudaError("Kernel execution failed");
	checkCudaErrors (cudaDeviceSynchronize());
	// re-read same images. shift them, update gpu_images and convert to gpu_clt;
	setImgBuffersShifted(
			is_bayer,   // int is_bayer,
			image_dx,   // int image_dx,
			image_dy); // int image_dy)
	convert_direct<<<grid_tp,threads_tp>>>( // called with a single block, CONVERT_DIRECT_INDEXING_THREADS threads
			m_tpParams.num_cams,      // int                num_cams,           // actual number of cameras
			m_tpParams.num_colors,    // int                num_colors,         // actual number of colors: 3 for RGB, 1 for LWIR/mono
			fgpu_kernel_offsets,      // struct CltExtra ** gpu_kernel_offsets,
			m_gpu_kernels,            // float           ** gpu_kernels,
			m_gpu_images,             // float           ** gpu_images,
			m_gpu_ftasks,             // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
			m_gpu_clt,     //******   // float           ** gpu_clt,            // [num_cams][TILESY][TILESX][num_colors][DTT_SIZE*DTT_SIZE]
			dstride/sizeof(float),    // size_t             dstride, // for gpu_images
			m_tpParams.tp_tasks_size, // int                num_tiles) // number of tiles in task
			0,                        // int                lpf_mask)            // apply lpf to colors : bit 0 - red, bit 1 - blue, bit2 - green
			m_tpParams.img_width,     // int                woi_width,
			m_tpParams.img_height,    // int                woi_height,
			m_tpParams.kernels_hor,   // int                kernels_hor,
			m_tpParams.kernels_vert,  //,          // int                kernels_vert);
			m_gpu_active_tiles,       // int *              gpu_active_tiles,      // pointer to the calculated number of non-zero tiles
			m_gpu_num_active,         // int *              pnum_active_tiles);  //  indices to gpu_tasks
			m_tpParams.tilesx);       // int                tilesx)
	getLastCudaError("Kernel execution failed");
	checkCudaErrors(cudaDeviceSynchronize());
    StopWatchInterface *timerINTERSELF = 0;
    sdkCreateTimer(&timerINTERSELF);
    int num_corrs{}; // will get data from the gpu memory
    for (int i = i0; i < numIterations; i++)   {
    	if (i == 0){
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerINTERSELF);
    		sdkStartTimer(&timerINTERSELF);
    	}
    	correlate2D_inter<<<1,1>>>(             // only results in TD
    			m_tpParams.num_cams,            // int                num_cams,           // actual number of cameras
				sel_sensors,                    // int               sel_sensors,
				m_gpu_clt,                      // float          ** gpu_clt,            // [num_cams] ->[TILES-Y][TILES-X][colors][DTT_SIZE*DTT_SIZE]
				gpu_clt_ref, // *********       // float          ** gpu_clt_ref,        // [num_cams] ->[TILES-Y][TILES-X][colors][DTT_SIZE*DTT_SIZE]
				m_tpParams.num_colors,          // int               colors,             // number of colors (3/1)
				m_tpParams.color_weights[0],    // 0.25,  // float             scale0,             // scale for R
				m_tpParams.color_weights[1],    // 0.25,  // float             scale1,             // scale for B
				m_tpParams.color_weights[2],    // 0.5,   // float             scale2,             // scale for G
				m_gpu_ftasks,                   // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
				m_tpParams.tp_tasks_size,       // int               num_tiles) // number of tiles in task
				m_tpParams.tilesx,              // int               tilesx,             // number of tile rows
				m_gpu_corr_indices,             // int             * gpu_corr_indices,   // packed tile+pair
				m_gpu_num_corr_tiles,           // int             * pnum_corr_tiles,    // pointer to a number of correlation tiles to process
				dstride_corr_td/sizeof(float),  // const size_t      corr_stride,        // in floats
				m_gpu_corrs_td);                // float           * gpu_corrs);         // correlation output data
        getLastCudaError("Kernel failure:correlate2D_inter");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("correlate2D_inter-TD pass: %d\n",i);
    	checkCudaErrors(cudaMemcpy(
    			&num_corrs,
				m_gpu_num_corr_tiles,
    			sizeof(int),
    			cudaMemcpyDeviceToHost));
    	checkCudaErrors(cudaDeviceSynchronize());
       	corr2D_normalize<<<1,1>>>(
        			num_corrs, //tp_task_size,                 // int               num_corr_tiles,     // number of correlation tiles to process
    				dstride_corr_td/sizeof(float),             // const size_t      corr_stride_td,     // in floats
					m_gpu_corrs_td,                            // float           * gpu_corrs_td,       // correlation tiles in transform domain
    				(float *) 0, // corr_weights,              // float           * corr_weights,       // null or per-tile weight (fat_zero2 will be divided by it)
    				dstride_corr/sizeof(float),                // const size_t      corr_stride,        // in floats
					m_gpu_corrs,                               // float           * gpu_corrs,          // correlation output data (pixel domain)
					m_tpParams.fat_zero * m_tpParams.fat_zero, // float             fat_zero2,           // here - absolute
					m_tpParams.corr_out_rad);                  // int               corr_radius);        // radius of the output correlation (7 for 15x15)
     	getLastCudaError("Kernel failure:corr2D_normalize");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("corr2D_normalize pass: %d\n",i);
    }
    sdkStopTimer(&timerINTERSELF);
    float avgTimeINTERSELF = (float)sdkGetTimerValue(&timerINTERSELF) / (float)numIterations;
    sdkDeleteTimer(&timerINTERSELF);
    printf("Average CORR-TD and companions run time =%f ms, num cor tiles (old) = %d\n",  avgTimeINTERSELF, num_corrs);
    saveInterCorrFile(
    		m_tpPaths.result_interscene_td,      // const char * path,   // "clt/aux_interscene-TD.raw"     m_tpPaths.result_interscene_td
			"interscene phase correlation",      // const char * prompt, // "interscene phase correlation"
			num_corrs,                           // int          num_corrs,
			num_corr_indices,                    // int          num_corr_indices,
			m_gpu_corrs_td,                      // float      * gpu_corrs_td,
			m_gpu_corr_indices,                  // int        * gpu_corr_indices,
			num_sel_sensors);                    // int          num_sel_sensors);
    saveInterCorrIndicesFile(
    		m_tpPaths.result_interscene_indices, // const char * path,   //  "clt/aux_inter-indices.raw" m_tpPaths.result_interscene_indices
			"interscene indices",                // const char * prompt, //  "interscene indices"
			num_corr_indices,                    // int          num_corr_indices,
			m_gpu_corr_indices,                  // int *        gpu_corr_indices,
			num_sel_sensors);                    // int          num_sel_sensors)
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
    	gfree(gpu_clt_ref_h[ncam]);
    }
    gfree(gpu_clt_ref);
}

void TpHostGpu::saveIntraCorrFile(
		const char * path,
		const char * prompt,
		int          num_corrs,
		int          num_corr_indices,
		float *      gpu_corrs,
		int *        gpu_corr_indices,
		int          num_sel_sensors) { // only for interscene
	if (!path)	return;
	int rslt_corr_length =   num_corrs * m_tpParams.corr_length;
	int corr_img_size =      num_corr_indices * 16 * 16; // NAN
//	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	float * cpu_corr = (float *)malloc(rslt_corr_length * sizeof(float));
	int *   cpu_corr_indices = (int *) malloc(num_corr_indices * sizeof(int));
	checkCudaErrors(cudaMemcpy2D(
			cpu_corr,
			m_tpParams.corr_length * sizeof(float),
			gpu_corrs,
			dstride_corr,
			m_tpParams.corr_length * sizeof(float),
			num_corrs,
			cudaMemcpyDeviceToHost));
	checkCudaErrors(cudaMemcpy(
			cpu_corr_indices,
			gpu_corr_indices,
			num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));
	float * corr_img = 	getCorrImg(
			corr_img_size,    // int     corr_img_size,
			num_corr_indices, //int     num_corr_indices,
			cpu_corr_indices, // int *   cpu_corr_indices,
			cpu_corr,         // float * cpu_corr,
			num_sel_sensors); // int     num_sel_sensors)
    printf("Writing %s to %s, width = %d, height=%d, slices=%d, length=%ld bytes\n",
    		prompt, path, (m_tpParams.tilesx * 16),(m_tpParams.tilesya * 16), m_tpParams.num_pairs, (corr_img_size * sizeof(float)) ) ;
    writeFloatsToFile(
    		corr_img,       // float *       data, // allocated array
			corr_img_size,  // int           size, // length in elements
			path);          // 			   const char *  path) // file path
    free (cpu_corr);
    free (cpu_corr_indices);
    free (corr_img);
}

/*

    printf("Writing interscene phase correlation data to %s, width = %d, height=%d, slices=%d, length=%ld bytes\n",
    		tpPaths.result_inter_td_norm_file, (TILESX*16),(TILESYA*16), num_pairs_inter, (corr_img_size * sizeof(float)) ) ;
    writeFloatsToFile(
    		corr_img,                  // float *       data, // allocated array
			corr_img_size,             // int           size, // length in elements
			tpPaths.result_inter_td_norm_file); // 			   const char *  path) // file path
 */

float * TpHostGpu::getCorrImg(
		int     corr_img_size,
		int     num_corr_indices,
		int *   cpu_corr_indices,
		float * cpu_corr,
		int     num_sel_sensors){
	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	for (int i = 0; i < corr_img_size; i++){
		corr_img[i] = NAN;
	}
    for (int ict = 0; ict < num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){ // Was not here - only for interscene. Will it hurt?
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	int src_offs0 = ict * m_tpParams.corr_length;
    	int dst_offs0 = cpair * (m_tpParams.num_tiles * 16 * 16) +  (ty * 16 * m_tpParams.tilesx * 16) + (tx * 16);
    	for (int iy = 0; iy < m_tpParams.corr_size; iy++){
    		int src_offs = src_offs0 + iy * m_tpParams.corr_size; // ict * num_pairs * corr_size * corr_size;
    		int dst_offs = dst_offs0 + iy * (m_tpParams.tilesx * 16);
    		for (int ix = 0; ix < m_tpParams.corr_size; ix++){
    			corr_img[dst_offs++] = cpu_corr[src_offs++];
    		}
    	}
    }
	return corr_img;
}

float * TpHostGpu::getCorrTdImg(
		int     corr_img_size,
		int     num_corr_indices,
		int *   cpu_corr_indices,
		float * cpu_corr_td,
		int     num_sel_sensors){
	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	for (int i = 0; i < corr_img_size; i++){
		corr_img[i] = NAN;
	}
    for (int ict = 0; ict < num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){ // Was not here - only for interscene. Will it hurt?
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	int src_offs0 = ict * m_tpParams.corr_length;
    	int dst_offs0 = cpair * (m_tpParams.num_tiles * 16 * 16) +  (ty * 16 * m_tpParams.tilesx * 16) + (tx * 16);

    	for (int iquad = 0; iquad < 4; iquad ++) {
    		int iqy = (iquad >> 1) & 1;
    		int iqx = (iquad >> 0) & 1;
    		for (int iy = 0; iy < m_tpParams.dtt_size; iy++){
    			int src_offs = src_offs0 + iy * m_tpParams.dtt_size + iquad * m_tpParams.dtt_size * m_tpParams.dtt_size;
    			int dst_offs = dst_offs0 + (iy + m_tpParams.dtt_size * iqy)* (m_tpParams.tilesx * 16) + iqx * m_tpParams.dtt_size;
    			for (int ix = 0; ix < m_tpParams.dtt_size; ix++){
    				corr_img[dst_offs++] = cpu_corr_td[src_offs++];
    			}
    		}
    	}
    }
	return corr_img;
}



void TpHostGpu::saveInterCorrFile(
		const char * path,   // "clt/aux_interscene-TD.raw"
		const char * prompt, // "interscene phase correlation"
		int          num_corrs,
		int          num_corr_indices,
		float      * gpu_corrs_td,
		int        * gpu_corr_indices,
		int          num_sel_sensors){
	if (!path)	return;
	int corr_img_size =    num_corr_indices * 16 * 16; // NAN
 	int rslt_corr_size_td =   num_corrs *  m_tpParams.dtt_size2 * m_tpParams.dtt_size2;
 	float * cpu_corr_td = (float *)malloc(rslt_corr_size_td * sizeof(float));
 	int     dtile_bytes = (m_tpParams.dtt_size2*m_tpParams.dtt_size2) * sizeof(float);
	int *   cpu_corr_indices = (int *) malloc(num_corr_indices * sizeof(int));

    checkCudaErrors(cudaMemcpy2D(
    		cpu_corr_td,
			dtile_bytes,
			gpu_corrs_td,
			dstride_corr_td,
			dtile_bytes,
			num_corrs,
			cudaMemcpyDeviceToHost));
    checkCudaErrors(cudaMemcpy(
    		cpu_corr_indices,
			gpu_corr_indices,
			num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));
	float * corr_img = 	getCorrTdImg(
			corr_img_size,    // int     corr_img_size,
			num_corr_indices, //int     num_corr_indices,
			cpu_corr_indices, // int *   cpu_corr_indices,
			gpu_corrs_td,     // float * cpu_corr,
			num_sel_sensors); // int     num_sel_sensors)
    printf("Writing %s TD data to %s\n", prompt, path);
    writeFloatsToFile(
    		corr_img,        // float *       data, // allocated array
			corr_img_size,   // int           size, // length in elements
			path);           // 			   const char *  path) // file path
    free (cpu_corr_indices);
	free (corr_img);
    free (cpu_corr_td);
}

void TpHostGpu::saveInterCorrIndicesFile(
		const char * path,   //  "clt/aux_inter-indices.raw"
		const char * prompt, //  "interscene indices"
		int          num_corr_indices,
		int *        gpu_corr_indices,
		int          num_sel_sensors){
	if (!path)	return;
	int *   cpu_corr_indices = (int *) malloc(num_corr_indices * sizeof(int));
    checkCudaErrors(cudaMemcpy(
    		cpu_corr_indices,
			gpu_corr_indices,
			num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));
    int corr_index_img_length = m_tpParams.tilesx * m_tpParams.tilesy * (num_sel_sensors+1) ;
    float *corr_index_img =  (float *)malloc(corr_index_img_length * sizeof(float));
    for (int i = 0; i < corr_index_img_length; i++){
    	corr_index_img[i] = NAN;
    }
    for (int ict = 0; ict < num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	corr_index_img[cpair * m_tpParams.tilesx * m_tpParams.tilesy + m_tpParams.tilesx * ty + tx] =  ict; // cpu_corr_indices[ict];
    }
    printf("Writing %s to %s\n",prompt,path);
    writeFloatsToFile(
    		corr_index_img,                  // float *       data, // allocated array
			corr_index_img_length,           // int           size, // length in elements
			"clt/aux_inter-indices.raw"); // 			   const char *  path) // file path
    free (corr_index_img);
    free (cpu_corr_indices);
}



void TpHostGpu::hfree(float *              p) {if (p) free(p); p = {};}
void TpHostGpu::hfree(struct CltExtra *    p) {if (p) free(p); p = {};}
void TpHostGpu::gfree(float *              p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(int *                p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct CltExtra *    p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct gc *          p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct corr_vector * p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct trot_deriv  * p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(float **             p) {if (p) checkCudaErrors(cudaFree(p));p = {};}


